/*
 *  XT->USB converter for IBM Model F 83-key XT keyboard.
 *  
 *  Arduino MICRO board.
 *  
 *  Supports full N-key rollover.
 *  
 *  The firmware does NOT support the legacy/boot protocol,
 *  so the keyboard will not be recognised by the BIOS, so cannot
 *  be used in BIOS-only environments.
 *  
 *  Once the OS brings up the USB HID interface, the keyboard
 *  becomes fully functional.
 *  
 *  This firmware is for the 83-key Model F keyboards only.
 *  Because this model has no LEDs, the firmware does not
 *  implement any LED controls.
 *
 *  Copyright (c) 2016, Sean Jensen.
 *  
 *  See file LICENSE for license terms.
 *
 */

#include "NKRO.h"

/*
 * XT_MATRIX is a bit-vector to hold the state of the keyboard.
 * 83 keys plus NULL value = 84 bits. 84 bits == 10 bytes + 4 bits, so we
 * need 11 bytes to store the vector.
 * 
 * This information is not actually used in this implementation, but will
 * be useful if state-aware keypress processing is ever required,
 * for example, assigning HID usage codes to arbitrary key-chords.
 * 
 */

unsigned char XT_MATRIX[11] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00
} ;

/*
 * Bit-banging macros for accessing bit-vectors, with
 * byte and bit ordering as expected by the USB protocol.
 * 
 */
 
#define SETQ(TBL,C) (TBL[C >> 3] & (1 << (C & 0x07))) 
#define SET(TBL,C) (TBL[C >> 3] |= (1 << (C & 0x07))) 
#define UNSET(TBL,C) (TBL[C >> 3] &= ~(1 << (C & 0x07)))

/*
 * Bitmask macros for parsing XT scancodes.
 * 
 */

#define KEY_BREAK 0x80
#define KEY_CODE 0x7F

/*
 * Keyboard's CLOCK is mapped to an external interrupt
 * so it should be connected to PIN 3 or PIN 7 on
 * the Arduino MICRO.
 * 
 */

#define PIN_CLOCK 3
#define PIN_DATA 4


/*
 *  READBIT() is the interrupt service routine, which is invoked
 *  every time the CLOCK rises. See the IBM XT Technical
 *  Reference (1986) for details of the CLOCK/DATA protocol.
 *  
 */

#define KEY_READY 0x01
#define KEY_NEXTBIT 0x100

volatile unsigned int RX ;    // Shift register to collect bits from keyboard DATA line.
volatile unsigned char CX ;   // Register to return completed scancode

void READBIT() {
  
  if (RX & KEY_READY) {
  
    /*
     * The start bit acts as a sentinel. When it reaches bit 0
     * we know we have read all the available bits. So now
     * having  hit the rising edge after pulse 9, we have
     * collected a complete scancode.
     * 
     */
   
    RX >>= 1 ;                             // Discard the sentinel start-bit.
    CX = (unsigned char) RX ;              // Make the scancode available.
    RX = 0 ;                               // Reset RX

  } else {

    /*
     * We are still collecting bits...
     */

    RX >>= 1 ;                             // Shift RX right

    if (digitalRead(PIN_DATA) == HIGH)     // Set the incoming bit if DATA line is active
      RX |= KEY_NEXTBIT ;
   
  }
 
}

/*
 *  UPDATE_MATRIX() parses the scancode and
 *  sets or unsets the appropriate bit in the
 *  XT_MATRIX state vector.
 *  
 */

void UPDATE_MATRIX(unsigned char K) {

  /*
   *  The SET and UNSET macros potentially evaluate
   *  C twice, hence the local variable here.
   *
   */

  unsigned char C = (K & KEY_CODE) ;

  if (K & KEY_BREAK) 
    UNSET(XT_MATRIX, C) ;
  else
    SET(XT_MATRIX, C) ;

} 

/*
 *  UPDATE_REPORT() parses the scancode and
 *  sets or unsets the appropriate bit in the
 *  XT_REPORT HID usage state vector.
 *  
 */

void UPDATE_REPORT(unsigned char K) {

  /*
   *  The SET and UNSET macros potentially evaluate
   *  C twice, hence the local variable here.
   *
   */

  unsigned char C = (K & KEY_CODE) ;

  if (K & KEY_BREAK) 
    UNSET(XT_REPORT, XT_HID_ID[C]) ;
  else
    SET(XT_REPORT, XT_HID_ID[C]) ;

} 

/*
 *  KB_ERROR() is the error handler. It
 *  reports the appropriate HID error usage
 *  code to the USB host. The error codes are:
 *    0x01 --> Overflow error
 *    0x02 --> POST error
 *    0x03 --> Unknown error
 *  
 */

void KB_ERROR(unsigned char e) {

  int i;

  for (i = 0; i < sizeof XT_MATRIX; i++)
    XT_MATRIX[i] = 0x00 ;
    
  for (i = 0; i < sizeof XT_REPORT; i++)
    XT_REPORT[i] = 0x00 ;
  
  SET(XT_REPORT, e) ;

  SEND() ;

}


/*
 *  Firmware entry-point.
 *  
 */
 
unsigned char SCANCODE ;
unsigned char LAST_SCANCODE ;

void setup() {

  /*
   * The Report Descriptor seems to need to be attached quickly.
   * 
   */
 
  ATTACH_SUBDESCRIPTOR() ;
 
  /*
   * Give ourselves some breathing space. This is the opportunity
   * to upload new firmware in case we disappear into
   * an unresponsive infinite loop.
   * 
  */

  delay(2000) ;

  /*
   *  Initialise the collector and return registers.
   *
   */

  RX = 0 ;
  CX = 0 ;

  /*
   *  Give control of the DATA line to the keyboard.
   *
   */

  pinMode(PIN_DATA, INPUT) ;

  /*
   * Send a RESET signal to the keyboard:
   * Pull the CLOCK down for a minimum of
   * 12.5 ms (according to the specs), and
   * then release.
   *
   */

  pinMode(PIN_CLOCK,OUTPUT) ;      // Take control of the CLOCK line
  digitalWrite(PIN_CLOCK,LOW) ;    // Take CLOCK LOW
  delay(20) ;                      // Keep it there for >12.5ms
  pinMode(PIN_CLOCK,INPUT);        // Return control of CLOCK to the keyboard

  /*
   * Keyboard feeds us data on the rising edge of the clock,
   * which we catch with an interrupt service routine, READBIT.
   *
   */

  attachInterrupt(digitalPinToInterrupt(PIN_CLOCK),READBIT,RISING) ;

  /*
   *  After a RESET, the keyboard performs a Basic
   *  Assurance Test (BAT). This can take up to
   *  500ms, and returns one of three status codes:
   *  
   *    0xAA  -->  BAT passed
   *    0xFC  -->  Failure Code (BAT failed)
   *    0xFF  -->  Overrun. Keyboard's internal buffer is overflowing.
   *    
   *  The HID specifications provide equivalent error usage codes
   *  which we despatch here, and then die.
   *  
   */

  delay(600) ;                              // Wait for the BAT to finish

  SCANCODE = CX ;                           // Retrieve the status byte
 
  switch (SCANCODE) {                       // Handle the result:
    case 0xAA : break ;                     //   BAT passed
    case 0xFC : KB_ERROR(0x02) ; exit(0) ;  //   BAT Failure Code (send HID code 0x02)
    case 0xFF : KB_ERROR(0x01) ; exit(0) ;  //   Overrun Error (send HID code 0x01)
    default   : KB_ERROR(0x03) ; exit(0) ;  //   Unknown Error (send HID code 0x03)
  }

  /*
   *  Prepare to enter the event loop...
   *  
   */

  LAST_SCANCODE = 0 ;
  CX = 0;

}

/*
 *  The event loop continuously polls READBIT's CX register.
 *  If it receives a scancode that differs from the previoua scancode
 *  it updates its internal state accordingly, translates the scancode
 *  into a HID usage code, updates the HID Report, and despatches it to the USB host.
 *  
 */

void loop() {

  SCANCODE = CX ;                         // Poll the READBIT collector

  if (SCANCODE != LAST_SCANCODE) {        // Do we have a new scancode?

    /*
     *  The keyboard can be told to stop sending
     *  data by holding the DATA line low for as long
     *  as necessary. Any scancodes generated in  
     *  the meantime will be held by the keyboard
     *  in its own internal 16-key buffer.
     *  
     *  While processing a complete
     *  scancode, we stop the keyboard sending data
     *  in case we miss any while we are busy here.
     *  
     */
     
    pinMode(PIN_DATA,OUTPUT) ;              // Take control of the DATA line
    digitalWrite(PIN_DATA,LOW) ;            // Take DATA LOW

    if (SCANCODE > 0xD6) {                  // Upper bound check. There are only 83 x 2 scancodes.
                                            // Oops, something is very wrong...
     if (SCANCODE == 0xFF)                  
        KB_ERROR(0x01) ;                    //   Internal keyboard buffer overflow.
     else
        KB_ERROR(0x03) ;                    //   Should never get here.
       
    } else {                                // We have a scancode
      
      UPDATE_MATRIX(SCANCODE) ;             //   Update our internal state
      UPDATE_REPORT(SCANCODE) ;             //   Update the HID Report
      SEND() ;                              //   Send the Report up to the USB host.
     
    }

    pinMode(PIN_DATA,INPUT) ;               // Give control of the DATA line back to the keyboard
  }

  LAST_SCANCODE = SCANCODE ;                // We now have a new previous scancode

}
